import { Injectable } from '@nestjs/common';
import { InjectEntityManager, InjectRepository } from '@nestjs/typeorm';

import { EntityManager, Repository } from 'typeorm';
import { Pagination } from '~/helper/paginate/pagination';
import { BusinessException } from '~/common/exceptions/biz.exception';
import { ErrorEnum } from '~/constants/error-code.constant';
import { paginate } from '~/helper/paginate';
import { Storage } from '~/modules/tools/storage/storage.entity';
import {
  MaterialsInOutQueryDto,
  MaterialsInOutDto,
  MaterialsInOutUpdateDto
} from './materials_in_out.dto';
import { MaterialsInOutEntity } from './materials_in_out.entity';
import { fieldSearch } from '~/shared/database/field-search';
import { ParamConfigEntity } from '~/modules/system/param-config/param-config.entity';
import { MaterialsInOrOutEnum, ParamConfigEnum } from '~/constants/enum';
import { MaterialsInventoryEntity } from '../materials_inventory.entity';
import { MaterialsInventoryService } from '../materials_inventory.service';
import { isDefined } from 'class-validator';

@Injectable()
export class MaterialsInOutService {
  constructor(
    @InjectEntityManager() private entityManager: EntityManager,
    @InjectRepository(MaterialsInOutEntity)
    private materialsInOutRepository: Repository<MaterialsInOutEntity>,
    @InjectRepository(Storage)
    private storageRepository: Repository<Storage>,
    @InjectRepository(ParamConfigEntity)
    private paramConfigRepository: Repository<ParamConfigEntity>,
    private materialsInventoryService: MaterialsInventoryService
  ) {}
  /**
   * 查询所有出入库记录
   */
  async findAll({
    page,
    pageSize,
    product: productName,
    projectId,
    isCreateOut,
    ...ext
  }: MaterialsInOutQueryDto): Promise<Pagination<MaterialsInOutEntity>> {
    const sqb = this.materialsInOutRepository
      .createQueryBuilder('materialsInOut')
      .leftJoin('materialsInOut.files', 'files')
      .leftJoin('materialsInOut.project', 'project')
      .leftJoin('materialsInOut.product', 'product')
      .leftJoin('product.unit', 'unit')
      .leftJoin('product.company', 'company')
      .addSelect([
        'files.id',
        'files.path',
        'project.name',
        'product.name',
        'unit.label',
        'company.name'
      ])
      .where(fieldSearch(ext))
      .andWhere('materialsInOut.isDelete = 0')
      .addOrderBy('materialsInOut.createdAt', 'DESC');

    if (productName) {
      sqb.andWhere('product.name like :productName', { productName: `%${productName}%` });
    }

    if (projectId) {
      sqb.andWhere('project.id = :projectId', { projectId });
    }

    if (isCreateOut) {
      sqb.andWhere('materialsInOut.inOrOut = 0');
    }
    const pageData = await paginate<MaterialsInOutEntity>(sqb, {
      page,
      pageSize
    });
    return pageData;
  }

  /**
   * 新增
   */
  async create(dto: MaterialsInOutDto): Promise<void> {
    let { inOrOut, inventoryNumber, projectId } = dto;
    if (Object.is(inOrOut, MaterialsInOrOutEnum.In)) {
      // 入库
      inventoryNumber = await this.generateInventoryNumber();
    } else {
      // 出库
      const inRecord = await this.materialsInOutRepository.findOne({
        where: {
          inventoryNumber
        }
      });
      const { productId } = inRecord;
      dto.productId = productId;
    }

    await this.entityManager.transaction(async manager => {
      // 1.生成出入库记录
      const { productId, quantity, unitPrice } = await manager.save(MaterialsInOutEntity, {
        ...this.materialsInOutRepository.create(dto),
        inventoryNumber
      });
      // 2.更新增减库存
      await (
        Object.is(inOrOut, MaterialsInOrOutEnum.In)
          ? this.materialsInventoryService.inInventory
          : this.materialsInventoryService.outInventory
      )({ productId, quantity, unitPrice, projectId }, manager);
    });
  }

  /**
   * 更新
   */
  async update(id: number, { fileIds, ...data }: Partial<MaterialsInOutUpdateDto>): Promise<void> {
    await this.entityManager.transaction(async manager => {
      const entity = await manager.findOne(MaterialsInOutEntity, {
        where: {
          id
        },
        lock: { mode: 'pessimistic_write' }
      });
      await manager.update(MaterialsInOutEntity, id, {
        ...data
      });
      let changedQuantity = 0;
      if (isDefined(data.quantity) && entity.quantity !== data.quantity) {
        if (entity.inOrOut === MaterialsInOrOutEnum.In) {
          // 入库减少等于出库
          if (data.quantity - entity.quantity < 0) {
            data.inOrOut = MaterialsInOrOutEnum.Out;
          } else {
            // 入库增多等于入库
            data.inOrOut = MaterialsInOrOutEnum.In;
          }
        } else {
          // 出库减少等于入库
          if (data.quantity - entity.quantity < 0) {
            data.inOrOut = MaterialsInOrOutEnum.In;
          } else {
            // 出库增多等于出库
            data.inOrOut = MaterialsInOrOutEnum.Out;
          }
        }
        changedQuantity = Math.abs(data.quantity - entity.quantity);
      }
      // 2.更新增减库存
      if (changedQuantity !== 0) {
        await (
          Object.is(data.inOrOut, MaterialsInOrOutEnum.In)
            ? this.materialsInventoryService.inInventory
            : this.materialsInventoryService.outInventory
        )(
          {
            productId: entity.productId,
            quantity: Math.abs(changedQuantity),
            unitPrice: undefined,
            projectId: entity.projectId
          },
          manager
        );
      }
      if (fileIds?.length) {
        const count = await this.storageRepository
          .createQueryBuilder('storage')
          .where('storage.id in(:fileIds)', { fileIds })
          .getCount();
        if (count !== fileIds?.length) {
          throw new BusinessException(ErrorEnum.STORAGE_NOT_FOUND);
        }
        // 附件要批量插入
        await manager
          .createQueryBuilder()
          .relation(MaterialsInOutEntity, 'files')
          .of(id)
          .add(fileIds);
      }
    });
  }

  /**
   * 删除
   */
  async delete(id: number): Promise<void> {
    await this.entityManager.transaction(async manager => {
      const entity = await manager.findOne(MaterialsInOutEntity, {
        where: {
          id,
          isDelete: 0
        },
        lock: { mode: 'pessimistic_write' }
      });
      if (!entity) {
        throw new BusinessException(ErrorEnum.MATERIALS_IN_OUT_NOT_FOUND);
      }

      // 更新库存
      await (
        Object.is(entity.inOrOut, MaterialsInOrOutEnum.In)
          ? this.materialsInventoryService.outInventory
          : this.materialsInventoryService.inInventory
      )(
        {
          productId: entity.productId,
          quantity: entity.quantity,
          unitPrice: undefined,
          projectId: entity.projectId
        },
        manager
      );
    });

    // 出入库比较重要，做逻辑删除
    await this.materialsInOutRepository.update(id, { isDelete: 1 });
  }

  /**
   * 获取单个出入库信息
   */
  async info(id: number) {
    const info = await this.materialsInOutRepository
      .createQueryBuilder('materialsInOut')
      .where({
        id
      })
      .andWhere('materialsInOut.isDelete = 0')
      .getOne();
    return info;
  }

  /**
   * 解除附件关联
   * @param id 出入库ID
   * @param fileIds 附件ID
   */
  async unlinkAttachments(id: number, fileIds: number[]) {
    await this.entityManager.transaction(async manager => {
      const materialsInOut = await this.materialsInOutRepository
        .createQueryBuilder('materialsInOut')
        .leftJoinAndSelect('materialsInOut.files', 'files')
        .where('materialsInOut.id = :id', { id })
        .getOne();
      const linkedFiles = materialsInOut.files
        .map(item => item.id)
        .filter(item => !fileIds.includes(item));
      // 附件要批量更新
      await manager
        .createQueryBuilder()
        .relation(MaterialsInOutEntity, 'files')
        .of(id)
        .addAndRemove(linkedFiles, materialsInOut.files);
    });
  }

  /**
   * 生成库存单号
   * @returns 库存单号
   */
  async generateInventoryNumber() {
    const prefix =
      (
        await this.paramConfigRepository.findOne({
          where: {
            key: ParamConfigEnum.MaterialsInOutPrefix
          }
        })
      )?.value || '';
    const lastMaterial = await this.materialsInOutRepository
      .createQueryBuilder('materialsInOut')
      .select(
        `MAX(CAST(REPLACE(materialsInOut.inventoryNumber, '${prefix}', '') AS UNSIGNED))`,
        'maxInventoryNumber'
      )
      .getRawOne();
    const lastNumber = lastMaterial.maxInventoryNumber
      ? parseInt(lastMaterial.maxInventoryNumber.replace(prefix, ''))
      : 0;
    const newNumber = lastNumber + 1 < 1000 ? 1000 : lastNumber + 1;
    return `${prefix}${newNumber}`;
  }
}
